from pyglet.gl import *
from math import *
from util import *

GRAVITY = -0.1
RESTITUTION = 0.5
WORLD_RADIUS = 15 # Radius of player for collision detection
IMAGE_RADIUS = 35 # Radius of player in pixels of original image
DEFAULT_JUMP_POWER = 5

# We need to subclass from object for the .pos setter to work.
class Player(object):
    def __init__(self, x, y):
        self.reset(x, y)
        self.radius = WORLD_RADIUS
        self.image = pyglet.resource.image('player.png').get_texture()
        self.image_scale = 0.5 * self.image.width / IMAGE_RADIUS
        self.deathSound = pyglet.resource.media('uh.wav', streaming=False)
        self.hitSound = pyglet.resource.media('hit.wav', streaming=False)
        self.curPlatform = None
        self.jumpPlatform = None

    def _getPos(self):
        return Vec2(self.x, self.y)

    def _setPos(self, vec):
        self.x = vec.x
        self.y = vec.y

    pos = property(_getPos, _setPos)
    
    def reset(self, x, y):
        self.isDead = False
        self.dying = False
        self.deadness = 0;
        self.x = x
        self.y = y
        self.velocity = Vec2(0, 0)
        self.orientation = 0
        self.rotation = 0
        self.canJump = False
        self.jumpPower = DEFAULT_JUMP_POWER
        self.isPoweringUp = False
        self.score = 0

    def die(self):
        if not self.dying:
            self.velocity.y *= 3
            self.dying = True
            self.deathSound.play()
            pyglet.clock.schedule_once(self._die, 0.5)

    def _die(self, dt):
        self.isDead = True
        
    
    def update(self, platform, floorHeight):
        if (self.dying):
            self.deadness += 1.0 / 30.0
        
    	self.canJump = False
    	
        if self.isPoweringUp:
            self.jumpPower = min(self.jumpPower + 0.1, 7)
    	
        self.x += self.velocity.x
        self.y += self.velocity.y + 0.5 * GRAVITY

        # We start at 279.
        if not self.dying:
            self.score = max(self.score, self.y - 279)
        
        self.velocity.y += GRAVITY
        
        self.orientation += self.rotation
        
        # Collisions
        
        if self.y - self.radius < floorHeight:
            self.y = self.radius + floorHeight
            self.collide(Vec2(0, 1), 0)
            self.die()
        
        if self.x - self.radius < 0:
            self.x = self.radius
            self.collide(Vec2(1, 0), 0)
        
        if self.x + self.radius > 420:
            self.x = 420 - self.radius
            self.collide(Vec2(-1, 0), 0)

        while platform:
            self.maybeCollideWith(platform)
            platform = platform.next

    def maybeCollideWith(self, platform):
        ends = platform.getEndpoints()
        closest = self.pos.closestPoint(ends[0], ends[1])
        offset = self.pos - closest
        dist = offset.length()
        penetration = self.radius + platform.width - dist

        if penetration >= 0:
            if platform.next and platform.next.next and \
                    platform.next.next.next and \
                    not platform.next.next.next.next:
                platform.next.next.next.makeFollowing()

            bounced = self.collide(offset.normalize(), penetration)
            if (not bounced):
                self.curPlatform = platform
            elif (bounced == 1):
                self.jumpPlatform = platform
    	
    # Calculate new velocity and rotation after collision
    # Lucky me: I get to simplify:
    #   * Collision always on circle's radius with normal towards centre
    #   * Can ignore the mass of the player because...
    #   * Can treat the other object as static
    # Returns 2 for "I bounced", 1 for "I can jump", 0 for "I landed".
    def collide(self, normal, penetration):
        self.x += normal.x * penetration;
        self.y += normal.y * penetration;

        bounced = 2
        
        approachSpeed = -self.velocity.dot(normal)
        
        if approachSpeed < 0:
            return
        
        restitution = RESTITUTION
        
        if normal.y > 0.5:
            self.canJump = True
            self.jumpDir = normal
            
            if approachSpeed < 1:
            	restitution = 0
                bounced = 0
            else:
                bounced = 1
        
        # Apply contact impulse along normal: only affects velocity because of simplifications
        
        normalImpulse = approachSpeed * (1 + restitution)

        self.velocity += normalImpulse * normal
        
        # Now move onto friction
        
        tangent = Vec2(-normal.y, normal.x)
        
        MoI = self.radius * self.radius * 0.5 * 0.25 # Moment of Inertia (ignoring mass)
        
        # Calculate current velocity of contact point, in direction of tangent
        tangentVelocity = tangent.dot(self.velocity) - self.radius * self.rotation
        
        # Calculate change in velocity at contact point for a unit impulse along tangent
        velocityChangePerImpulse = 1.0 + self.radius * self.radius / MoI
                                 # 1.0 is linear component
                                 # r*r/MoI is magic angular component
        
        # Calculate strength of friction impulse required to balance current velocity
        friction = -tangentVelocity / velocityChangePerImpulse
        
        # Apply friction impulse to velocity and rotation
        
        self.velocity += friction * tangent
        self.rotation -= friction * self.radius / MoI

        if bounced:
            self.hitSound.play()

        return bounced
    
    def startPowerUp(self):
        self.isPoweringUp = True

    def stopPowerUp(self):
        self.isPoweringUp = False

    def jump(self):
        self.stopPowerUp()
        if not self.canJump:
            self.jumpPower = DEFAULT_JUMP_POWER
            return
        self.velocity += self.jumpDir.normalize() * self.jumpPower
        self.jumpPower = DEFAULT_JUMP_POWER

    def draw(self):
        glPushMatrix()
        glTranslatef(self.x, self.y, 0)
        glScalef(self.radius * self.image_scale, self.radius * self.image_scale, 1)
        glRotatef(self.orientation * 180 / pi, 0, 0, 1)
        
        glColor3f(1,1-self.deadness, 1-self.deadness)
        
        self.image.blit(-1, -1, width=2, height=2)

        glPopMatrix()
